#include <avr/io.h>
#include <stdlib.h> // for NULL

#include "lights.h"

static unsigned char led_counters[N_LEDS];
static pattern_t *led_patterns[N_LEDS];

static unsigned char fibonacci[8] = {
	0, 1, 2, 3, 5, 8, 13, 21,
};

static pattern_t boot_pattern[] = {
	{ 1, D_5 },
	{ 0, D_5 },
	{ 1, D_3 },
	{ 0, D_3 },
	{ 1, D_2 },
	{ 0, D_2 },
	{ 1, D_1 },
	{ 0, D_1 },
	{ 1, D_1 },
	{ 0, D_1 },
	{ 1, D_1 },
	{ 0, D_1 },
	{ 1, D_1 },
	{ 0, D_1 },
	{ 1, D_8 },
	{ 0, D_8 },
	PATTERN_END
};

static pattern_t pattern_num[] = {
	{ 0, D_5 },
	{ 1, D_1 }, /* 10 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  9 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  8 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  7 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  6 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  5 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  4 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  3 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  2 */
	{ 0, D_5 },
	{ 1, D_1 }, /*  1 */
	{ 0, D_13 },
	PATTERN_END
};

static pattern_t pattern_invnum[] = {
	{ 1, D_5 },
	{ 0, D_1 }, /* 10 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  9 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  8 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  7 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  6 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  5 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  4 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  3 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  2 */
	{ 1, D_5 },
	{ 0, D_1 }, /*  1 */
	{ 1, D_13 },
	PATTERN_END
};

pattern_t off_pattern[] = {
	{ 0, D_1 },
	PATTERN_END
};

static void led_set_mode(unsigned char n, unsigned char mode)
{
	if (n < N_PWMLEDS) {
		pwmled_set_mode(n, mode);
	} else if (n < N_LEDS) {
		gpio_set(n - N_PWMLEDS, mode);
	}
}

void led_set_pattern(unsigned char n, pattern_t *pattern)
{
	if (!pattern)
		pattern = off_pattern;

	led_patterns[n] = pattern;

	led_counters[n] = fibonacci[pattern->duration_fib];

	led_set_mode(n, pattern->mode);
}

void init_pattern()
{
	unsigned char i;

	for (i = 0; i < N_LEDS; i++)
		led_set_pattern(i, NULL);

	led_set_pattern(N_PWMLEDS+1, boot_pattern);
}

pattern_t *number_pattern(unsigned char num, unsigned char inv)
{
	if (num >= 10)
		num = 10;

	if (inv) {
		return pattern_invnum
			+ sizeof(pattern_invnum)/sizeof(pattern_t)
			- 2 - 2*num;
	} else {
		return pattern_num
			+ sizeof(pattern_num)/sizeof(pattern_t)
			- 2 - 2*num;
	}
}

static pattern_t *pattern_select(unsigned char n)
{
	switch(n) {
	case 0: return pwmled0_pattern_select();
	case 1: return pwmled1_pattern_select();
	case 2: return pwmled2_pattern_select();
	case 3: return status_led_pattern_select();
	case 4: return illumination_led_pattern_select();
	case 6: return laser_pattern_select();
	default: return NULL;
	}
}

void pattern_reload()
{
	unsigned char i;

	for (i = 0; i < N_LEDS; i++)
		led_set_pattern(i, pattern_select(i));
}

static void inline pattern_finished(unsigned char n)
{
	unsigned char i;

	led_patterns[n] = NULL;

	if (n < N_PWMLEDS) {
		for (i = 0; i < N_PWMLEDS; i++)
			if (led_patterns[i])
				return;

		/* all pwmleds finished; restart them */
		for (i = 0; i < N_PWMLEDS; i++)
			led_set_pattern(i, pattern_select(i));
	} else if (n == 3) {
		if (!led_patterns[4])
			led_set_pattern(4, pattern_select(4));
	} else if (n == 4) {
		if (!led_patterns[3])
			led_set_pattern(3, pattern_select(3));
	} else {
		led_set_pattern(n, pattern_select(n));
	}
}

void patterns_next_tick()
{
	unsigned char i;

	for (i = 0; i < N_LEDS; i++) {
		if (!led_patterns[i]) {
			pattern_finished(i);
			continue;
		}

		if (--led_counters[i] == 0) {
			pattern_t *p = led_patterns[i];
			p++;
			if (p->duration_fib == 0) { // END
				/* Keep the last state, wait for others */
				pattern_finished(i);
				continue;
			}
			led_set_pattern(i, p);
		}

	}
}

